from io import StringIO
from unittest.mock import patch

from django.core.management import call_command

from app_utils.testing import NoSocketsTestCase

PACKAGE_PATH = "package_monitor.management.commands"


@patch(PACKAGE_PATH + ".package_monitor_refresh.Distribution.objects.update_all")
class TestPurgeAll(NoSocketsTestCase):
    def test_can_purge_all_data(self, mock_update_all):
        mock_update_all.return_value = 0

        out = StringIO()
        call_command("package_monitor_refresh", stdout=out)
        self.assertTrue(mock_update_all.called)



from unittest.mock import patch

from django.test import RequestFactory, TestCase
from django.urls import reverse

from app_utils.testing import create_fake_user, json_response_to_python

from package_monitor import views

from .factories import DistributionFactory

MODULE_PATH_VIEWS = "package_monitor.views"
MODULE_PATH_MANAGERS = "package_monitor.managers"


@patch(MODULE_PATH_MANAGERS + ".PACKAGE_MONITOR_SHOW_ALL_PACKAGES", True)
@patch(MODULE_PATH_MANAGERS + ".PACKAGE_MONITOR_INCLUDE_PACKAGES", [])
class TestPackageList(TestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.user = create_fake_user(
            1001, "Bruce Wayne", permissions=["package_monitor.basic_access"]
        )
        cls.factory = RequestFactory()

    def test_index_view(self):
        # given
        request = self.factory.get(reverse("package_monitor:index"))
        request.user = self.user
        # when
        response = views.index(request)
        # then
        self.assertEqual(response.status_code, 200)

    def test_list_view_all(self):
        # given
        DistributionFactory(name="alpha")
        DistributionFactory(name="bravo")
        request = self.factory.get(reverse("package_monitor:package_list_data"))
        request.user = self.user
        # when
        response = views.package_list_data(request)
        # then
        self.assertEqual(response.status_code, 200)
        package_names = [x["name"] for x in json_response_to_python(response)]
        self.assertListEqual(package_names, ["alpha", "bravo"])

    def test_list_view_outdated(self):
        # given
        DistributionFactory(name="alpha")
        DistributionFactory(
            name="bravo",
            is_outdated=True,
            installed_version="1.0.0",
            latest_version="1.1.0",
        )
        request = self.factory.get(
            reverse("package_monitor:package_list_data"), data={"filter": "outdated"}
        )
        request.user = self.user
        # when
        response = views.package_list_data(request)
        # then
        self.assertEqual(response.status_code, 200)
        package_names = [x["name"] for x in json_response_to_python(response)]
        self.assertListEqual(package_names, ["bravo"])

    def test_list_view_current(self):
        # given
        DistributionFactory(name="alpha")
        DistributionFactory(
            name="bravo",
            is_outdated=True,
            installed_version="1.0.0",
            latest_version="1.1.0",
        )
        request = self.factory.get(
            reverse("package_monitor:package_list_data"), data={"filter": "current"}
        )
        request.user = self.user
        # when
        response = views.package_list_data(request)
        # then
        self.assertEqual(response.status_code, 200)
        package_names = [x["name"] for x in json_response_to_python(response)]
        self.assertListEqual(package_names, ["alpha"])

    def test_list_view_unknown(self):
        # given
        DistributionFactory(name="alpha")
        DistributionFactory(name="charlie", is_outdated=None)
        request = self.factory.get(
            reverse("package_monitor:package_list_data"), data={"filter": "unknown"}
        )
        request.user = self.user
        # when
        response = views.package_list_data(request)
        # then
        self.assertEqual(response.status_code, 200)
        package_names = [x["name"] for x in json_response_to_python(response)]
        self.assertListEqual(package_names, ["charlie"])

    @patch(MODULE_PATH_VIEWS + ".Distribution.objects.update_all")
    def test_refresh_distributions_view(self, mock_update_all):
        mock_update_all.return_value = 1

        request = self.factory.get(reverse("package_monitor:refresh_distributions"))
        request.user = self.user
        response = views.refresh_distributions(request)
        self.assertEqual(response.status_code, 200)



"""Django app config for Package Monitor."""

from django.apps import AppConfig

from . import __version__


class PackageMonitorConfig(AppConfig):
    name = "package_monitor"
    label = "package_monitor"
    verbose_name = f"Package Monitor v{__version__}"



# Generated by Django 4.0.7 on 2022-08-30 16:47

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("package_monitor", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="distribution",
            name="is_editable",
            field=models.BooleanField(
                default=None,
                help_text="Package has been installed in editable mode, i.e. pip install -e",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="distribution",
            name="apps",
            field=models.JSONField(
                default=list,
                help_text="List of installed Django apps included in this package",
            ),
        ),
        migrations.AlterField(
            model_name="distribution",
            name="has_installed_apps",
            field=models.BooleanField(
                default=False,
                help_text="Whether this package has installed Django apps",
            ),
        ),
        migrations.AlterField(
            model_name="distribution",
            name="is_outdated",
            field=models.BooleanField(
                default=None,
                help_text="A package is outdated when there is a newer stable version available",
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name="distribution",
            name="used_by",
            field=models.JSONField(
                default=list,
                help_text="List of other distribution packages using this package",
            ),
        ),
    ]



import datetime as dt
from collections import namedtuple
from unittest.mock import patch

from django.test import TestCase

from package_monitor.core.schedule import is_notification_due

MODULE_PATH = "package_monitor.core.schedule"

UTC = dt.timezone.utc


class TestSchedule(TestCase):
    def test_should_report_due_when_due(self):
        X = namedtuple(
            "X", ["schedule_text", "max_delay", "last_report", "now", "result"]
        )
        cases = [
            X(
                schedule_text="every day at 10:00",
                max_delay=5400,
                last_report=dt.datetime(2024, 7, 10, 10, 0, 0, tzinfo=UTC),
                now=dt.datetime(2024, 7, 11, 10, 0, 1, tzinfo=UTC),
                result=True,
            ),
            X(
                schedule_text="every day at 10:00",
                max_delay=5400,
                last_report=dt.datetime(2024, 7, 11, 10, 5, 0, tzinfo=UTC),
                now=dt.datetime(2024, 7, 11, 10, 0, 1, tzinfo=UTC),
                result=False,
            ),
            X(
                schedule_text="every day at 10:00",
                max_delay=5400,
                last_report=None,
                now=dt.datetime(2024, 7, 11, 10, 0, 1, tzinfo=UTC),
                result=True,
            ),
            X(
                schedule_text="",
                max_delay=5400,
                last_report=dt.datetime(2024, 7, 11, 10, 5, 0, tzinfo=UTC),
                now=dt.datetime(2024, 7, 11, 10, 0, 1, tzinfo=UTC),
                result=True,
            ),
            X(
                schedule_text="every thursday at 10:00",
                max_delay=5400,
                last_report=dt.datetime(2024, 7, 11, 6, 10, 0, tzinfo=UTC),
                now=dt.datetime(2024, 7, 11, 10, 0, 1, tzinfo=UTC),
                result=True,
            ),
            X(
                schedule_text="every thursday at 10:00",
                max_delay=5400,
                last_report=dt.datetime(2024, 7, 11, 10, 10, 0, tzinfo=UTC),
                now=dt.datetime(2024, 7, 11, 10, 0, 1, tzinfo=UTC),
                result=False,
            ),
            X(
                schedule_text="every thursday at 10:00",
                max_delay=5400,
                last_report=dt.datetime(2024, 7, 11, 9, 0, 0, tzinfo=UTC),
                now=dt.datetime(2024, 7, 11, 10, 1, 0, tzinfo=UTC),
                result=True,
            ),
            X(
                schedule_text="every thursday at 10:00",
                max_delay=5400,
                last_report=dt.datetime(2024, 7, 11, 9, 0, 0, tzinfo=UTC),
                now=dt.datetime(2024, 7, 11, 11, 31, 0, tzinfo=UTC),
                result=False,
            ),
            X(
                schedule_text="every thursday at 10:00",
                max_delay=5400,
                last_report=dt.datetime(2024, 7, 11, 9, 0, 0, tzinfo=UTC),
                now=dt.datetime(2024, 7, 11, 9, 59, 0, tzinfo=UTC),
                result=False,
            ),
        ]
        for num, tc in enumerate(cases, start=1):
            with self.subTest(num=num):
                with patch(MODULE_PATH + ".now") as m:
                    m.return_value = tc.now
                    x = is_notification_due(
                        schedule_text=tc.schedule_text,
                        max_delay=tc.max_delay,
                        last_report=tc.last_report,
                    )
                self.assertIs(x, tc.result)



from .factories import (  # noqa: F401
    DistributionFactory,
    DistributionPackageFactory,
    MetadataDistributionStubFactory,
    PypiFactory,
    PypiReleaseFactory,
    make_packages,
)
from .stubs import DjangoAppConfigStub  # noqa: F401



"""Helpers to deal with the importlib metadata library."""

import json
import os
import sys
from typing import List, Optional

from importlib_metadata import Distribution as MetadataDistribution
from packaging.requirements import InvalidRequirement, Requirement

from django.apps import apps as django_apps


def is_distribution_editable(dist: MetadataDistribution) -> bool:
    """Determine if a distribution is an editable install?"""
    # method for new packages conforming with pep 660
    direct_url_json = dist.read_text("direct_url.json")
    if direct_url_json:
        direct_url = json.loads(direct_url_json)
        if "dir_info" in direct_url and direct_url["dir_info"].get("editable") is True:
            return True

    # method for old packages
    for path_item in sys.path:
        egg_link = os.path.join(path_item, dist.name + ".egg-link")
        if os.path.isfile(egg_link):
            return True

    return False


def identify_installed_django_apps(dist: MetadataDistribution) -> List[str]:
    """Identify installed Django apps in metadata distribution
    and return their app labels.
    """
    if not dist.files:
        return []
    if "apps.py" not in {path.name for path in dist.files}:
        return []
    found_apps = []
    for dist_file in _extract_files(dist, pattern="__init__.py"):
        for app in django_apps.get_app_configs():
            if not app.module:
                continue
            my_file = app.module.__file__
            if my_file and my_file.endswith(dist_file):
                found_apps.append(app.name)
                break
    return found_apps


def _extract_files(dist: Optional[MetadataDistribution], pattern: str) -> List[str]:
    """Extract file paths from a distribution which filename match a pattern."""
    if not dist or not dist.files:
        return []
    dist_files = [str(f) for f in dist.files if f.name == pattern]
    return dist_files


# def _determine_homepage_url(dist: MetadataDistribution) -> str:
#     if url := dist_metadata_value(dist, "Home-page"):
#         return url
#     values = dist.metadata.get_all("Project-URL")
#     while values:
#         k, v = [o.strip() for o in values.pop(0).split(",")]
#         if k.lower() == "homepage":
#             return v
#     return ""


def parse_requirements(dist: MetadataDistribution) -> List[Requirement]:
    """Parse requirements from a distribution and return them.
    Invalid requirements will be ignored.
    """
    if not dist.requires:
        return []
    requirements = []
    for requirement in dist.requires:
        try:
            requirements.append(Requirement(requirement))
        except InvalidRequirement:
            pass
    return requirements


def metadata_value(dist: MetadataDistribution, prop: str) -> str:
    """Metadata value from distribution or empty string.
    Note: metadata can contain multiple values for the same key.
    This method will return the first only!
    """
    if dist and (value := dist.metadata.get(prop)) and value != "UNKNOWN":
        return value
    return ""



"""Tasks for Package Monitor."""

from celery import chain, shared_task

from django.core.cache import cache
from django.utils.timezone import now

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from . import __title__
from .app_settings import (
    PACKAGE_MONITOR_NOTIFICATIONS_ENABLED,
    PACKAGE_MONITOR_NOTIFICATIONS_MAX_DELAY,
    PACKAGE_MONITOR_NOTIFICATIONS_REPEAT,
    PACKAGE_MONITOR_NOTIFICATIONS_SCHEDULE,
    PACKAGE_MONITOR_SHOW_EDITABLE_PACKAGES,
)
from .core import schedule
from .models import Distribution

logger = LoggerAddTag(get_extension_logger(__name__), __title__)
CACHE_KEY_LAST_REPORT = "package-monitor-notification-last-report"


@shared_task(time_limit=3600)
def update_distributions():
    """Run regular tasks."""
    if _should_send_notifications():
        chain(update_all_distributions.si(), send_update_notification.si()).delay()
    else:
        update_all_distributions.delay()


def _should_send_notifications() -> bool:
    if not PACKAGE_MONITOR_NOTIFICATIONS_ENABLED:
        return False
    last_report = cache.get(key=CACHE_KEY_LAST_REPORT)
    is_due = schedule.is_notification_due(
        schedule_text=PACKAGE_MONITOR_NOTIFICATIONS_SCHEDULE,
        max_delay=PACKAGE_MONITOR_NOTIFICATIONS_MAX_DELAY,
        last_report=last_report,
    )
    if not is_due:
        return False
    return True


@shared_task
def update_all_distributions():
    """Update all distributions."""
    Distribution.objects.update_all()


@shared_task
def send_update_notification(should_repeat: bool = False):
    """Send update notification to inform about new versions."""
    Distribution.objects.send_update_notification(
        show_editable=PACKAGE_MONITOR_SHOW_EDITABLE_PACKAGES,
        should_repeat=should_repeat or PACKAGE_MONITOR_NOTIFICATIONS_REPEAT,
    )
    cache.set(key=CACHE_KEY_LAST_REPORT, value=now(), timeout=None)



from collections import namedtuple
from unittest import mock

from packaging.specifiers import SpecifierSet

from app_utils.testing import NoSocketsTestCase

from package_monitor.models import Distribution

from .factories import DistributionFactory, DistributionPackageFactory, make_packages

MODULE_PATH = "package_monitor.managers"


@mock.patch(MODULE_PATH + ".update_packages_from_pypi", spec=True)
@mock.patch(MODULE_PATH + ".compile_package_requirements", spec=True)
@mock.patch(MODULE_PATH + ".gather_distribution_packages", spec=True)
class TestDistributionsUpdateAll(NoSocketsTestCase):
    def test_should_create_new_packages_from_scratch(
        self,
        mock_gather_distribution_packages,
        mock_compile_package_requirements,
        mock_update_packages_from_pypi,
    ):
        # given
        dist_alpha = DistributionPackageFactory(
            name="alpha",
            current="1.0.0",
            homepage_url="https://www.alpha.com",
            summary="alpha-description",
        )
        dist_bravo = DistributionPackageFactory(
            name="bravo",
            requires=["alpha>=1.0.0"],
            homepage_url="https://www.bravo.com",
        )
        packages = make_packages(dist_alpha, dist_bravo)
        packages["alpha"].apps = ["alpha_app"]
        mock_gather_distribution_packages.return_value = packages
        mock_compile_package_requirements.return_value = {
            "alpha": {"bravo": SpecifierSet(">=1.0.0")}
        }
        # when
        Distribution.objects.update_all()
        # then
        self.assertEqual(Distribution.objects.count(), 2)
        obj = Distribution.objects.get(name="alpha")
        self.assertEqual(obj.installed_version, "1.0.0")
        self.assertFalse(obj.is_outdated)
        self.assertTrue(obj.has_installed_apps)
        self.assertEqual(obj.website_url, "https://www.alpha.com")
        self.assertEqual(obj.description, "alpha-description")
        self.assertListEqual(obj.apps, ["alpha_app"])
        self.assertEqual(
            obj.used_by,
            [
                {
                    "name": "bravo",
                    "homepage_url": "https://www.bravo.com",
                    "requirements": [">=1.0.0"],
                }
            ],
        )

    def test_should_retain_package_name_with_capitals(
        self,
        mock_gather_distribution_packages,
        mock_compile_package_requirements,
        mock_update_packages_from_pypi,
    ):
        # given
        dist_alpha = DistributionPackageFactory(name="Alpha", current="1.0.0")
        mock_gather_distribution_packages.return_value = make_packages(dist_alpha)
        mock_compile_package_requirements.return_value = {}
        # when
        Distribution.objects.update_all()
        # then
        self.assertEqual(Distribution.objects.count(), 1)
        obj = Distribution.objects.first()
        self.assertEqual(obj.name, "Alpha")
        self.assertEqual(obj.installed_version, "1.0.0")
        self.assertFalse(obj.is_outdated)

    def test_should_update_existing_packages(
        self,
        mock_gather_distribution_packages,
        mock_compile_package_requirements,
        mock_update_packages_from_pypi,
    ):
        # given
        dist_alpha = DistributionPackageFactory(name="alpha", current="1.0.0")
        mock_gather_distribution_packages.return_value = make_packages(dist_alpha)
        mock_compile_package_requirements.return_value = {}
        DistributionFactory(name="alpha", installed_version="0.9.0")
        # when
        Distribution.objects.update_all()
        # then
        self.assertEqual(Distribution.objects.count(), 1)
        obj = Distribution.objects.get(name="alpha")
        self.assertEqual(obj.installed_version, "1.0.0")
        self.assertFalse(obj.is_outdated)

    def test_should_remove_stale_packages(
        self,
        mock_gather_distribution_packages,
        mock_compile_package_requirements,
        mock_update_packages_from_pypi,
    ):
        # given
        dist_alpha = DistributionPackageFactory(name="alpha", current="1.0.0")
        mock_gather_distribution_packages.return_value = make_packages(dist_alpha)
        mock_compile_package_requirements.return_value = {}
        DistributionFactory(name="alpha", installed_version="0.9.0")
        DistributionFactory(name="bravo", installed_version="1.0.0")
        # when
        Distribution.objects.update_all()
        # then
        self.assertEqual(Distribution.objects.count(), 1)
        obj = Distribution.objects.get(name="alpha")
        self.assertEqual(obj.installed_version, "1.0.0")
        self.assertFalse(obj.is_outdated)

    def test_should_set_is_outdated_to_none_when_no_pypi_infos(
        self,
        mock_gather_distribution_packages,
        mock_compile_package_requirements,
        mock_update_packages_from_pypi,
    ):
        # given
        dist_alpha = DistributionPackageFactory(name="alpha", current="")
        mock_gather_distribution_packages.return_value = make_packages(dist_alpha)
        mock_compile_package_requirements.return_value = {}
        DistributionFactory(name="alpha", installed_version="0.9.0")
        # when
        Distribution.objects.update_all()
        # then
        self.assertEqual(Distribution.objects.count(), 1)
        obj = Distribution.objects.get(name="alpha")
        self.assertEqual(obj.latest_version, "")
        self.assertIsNone(obj.is_outdated)

    def test_should_set_is_outdated_to_none_when_current_version_can_not_be_parsed(
        self,
        mock_gather_distribution_packages,
        mock_compile_package_requirements,
        mock_update_packages_from_pypi,
    ):
        # given
        dist_alpha = DistributionPackageFactory(name="alpha", current="2009r")
        packages = make_packages(dist_alpha)
        packages["alpha"].latest = ""
        mock_gather_distribution_packages.return_value = packages
        mock_compile_package_requirements.return_value = {}
        # when
        Distribution.objects.update_all()
        # then
        self.assertEqual(Distribution.objects.count(), 1)
        obj = Distribution.objects.get(name="alpha")
        self.assertEqual(obj.latest_version, "")
        self.assertIsNone(obj.is_outdated)


class TestDistributionFilterVisible(NoSocketsTestCase):
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_SHOW_ALL_PACKAGES", True)
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_SHOW_EDITABLE_PACKAGES", False)
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_INCLUDE_PACKAGES", [])
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_EXCLUDE_PACKAGES", [])
    def test_should_have_all_packages(self):
        # given
        obj_1 = DistributionFactory()
        obj_2 = DistributionFactory()
        # when
        result = Distribution.objects.filter_visible()
        # then
        self.assertEqual(result.names(), {obj_1.name, obj_2.name})

    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_SHOW_ALL_PACKAGES", False)
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_SHOW_EDITABLE_PACKAGES", False)
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_INCLUDE_PACKAGES", [])
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_EXCLUDE_PACKAGES", [])
    def test_should_have_apps_only(self):
        # given
        obj_1 = DistributionFactory(apps=["app_1"])
        DistributionFactory()
        # when
        result = Distribution.objects.filter_visible()
        # then
        self.assertEqual(result.names(), {obj_1.name})

    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_SHOW_ALL_PACKAGES", False)
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_SHOW_EDITABLE_PACKAGES", False)
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_INCLUDE_PACKAGES", ["include-me"])
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_EXCLUDE_PACKAGES", [])
    def test_should_have_apps_plus_included(self):
        # given
        obj_1 = DistributionFactory(apps=["app_1"])
        obj_2 = DistributionFactory(name="include-me")
        DistributionFactory()
        # when
        result = Distribution.objects.filter_visible()
        # then
        self.assertEqual(result.names(), {obj_1.name, obj_2.name})

    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_SHOW_ALL_PACKAGES", True)
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_SHOW_EDITABLE_PACKAGES", False)
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_INCLUDE_PACKAGES", [])
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_EXCLUDE_PACKAGES", ["exclude-me"])
    def test_should_have_all_packages_minus_excluded(self):
        # given
        obj_1 = DistributionFactory()
        DistributionFactory(name="exclude-me")
        # when
        result = Distribution.objects.filter_visible()
        # then
        self.assertEqual(result.names(), {obj_1.name})

    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_SHOW_ALL_PACKAGES", True)
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_SHOW_EDITABLE_PACKAGES", False)
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_INCLUDE_PACKAGES", [])
    @mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_EXCLUDE_PACKAGES", [])
    def test_should_have_all_packages_minus_editable(self):
        # given
        obj_1 = DistributionFactory()
        DistributionFactory(name="exclude-me", is_editable=True)
        # when
        result = Distribution.objects.filter_visible()
        # then
        self.assertEqual(result.names(), {obj_1.name})


@mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_SHOW_ALL_PACKAGES", True)
@mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_SHOW_EDITABLE_PACKAGES", False)
@mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_INCLUDE_PACKAGES", [])
@mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_EXCLUDE_PACKAGES", [])
class TestDistributionBuildInstallCommand(NoSocketsTestCase):
    def test_all_packages(self):
        # given
        DistributionFactory(name="alpha", latest_version="1.2.0")
        DistributionFactory(name="bravo", latest_version="2.1.0")
        # when
        result = Distribution.objects.order_by("name").build_install_command()
        # then
        self.assertEqual(result, "pip install alpha==1.2.0 bravo==2.1.0")

    def test_should_stay_within_max_line_length(self):
        # given
        DistributionFactory.create_batch(size=500)
        # when
        result = Distribution.objects.all().build_install_command()
        # then
        self.assertLessEqual(len(result), 4095)


class TestDistributionNotifyUpdates(NoSocketsTestCase):
    def test_can_notify_updates(self):
        X = namedtuple(
            "X",
            [
                "shouldNotify",
                "installed_version",
                "latest_version",
                "latest_notified_version",
                "is_editable",
                "show_editable",
                "should_repeat",
            ],
        )
        cases = [
            X(True, "1.0.0", "1.0.1", "", False, False, False),
            X(False, "1.0.0", "1.0.0", "", False, False, False),
            X(False, "1.0.0", "0.1.0", "", False, False, False),
            X(False, "1.0.0", "", "", False, False, False),
            X(False, "", "1.0.0", "", False, False, False),
            X(False, "1.0.0", "1.0.1", "1.0.1", False, False, False),
            X(True, "1.0.0", "1.0.2", "1.0.1", False, False, False),
            X(False, "1.0.0", "1.0.2", "", True, False, False),
            X(True, "1.0.0", "1.0.2", "", True, True, False),
            X(True, "1.0.0", "1.0.1", "1.0.1", False, False, True),
        ]
        for num, tc in enumerate(cases, 1):
            with self.subTest("test notifications", num=num):
                Distribution.objects.all().delete()
                dist = DistributionFactory(
                    installed_version=tc.installed_version,
                    latest_version=tc.latest_version,
                    latest_notified_version=tc.latest_notified_version,
                    is_editable=tc.is_editable,
                )
                with mock.patch(
                    MODULE_PATH + ".notify_admins", spec=True
                ) as notify_admins:
                    Distribution.objects.send_update_notification(
                        show_editable=tc.show_editable, should_repeat=tc.should_repeat
                    )
                    self.assertIs(tc.shouldNotify, notify_admins.called)
                    dist.refresh_from_db()
                    if tc.shouldNotify:
                        self.assertEqual(
                            dist.latest_version, dist.latest_notified_version
                        )
                    else:
                        self.assertEqual(
                            tc.latest_notified_version, dist.latest_notified_version
                        )



import unittest

from .stubs import PackageMetadataStub


class TestPackageMetadata(unittest.TestCase):
    def test_single(self):
        # given
        d = PackageMetadataStub()
        # when
        d["one"] = 1
        # then
        self.assertEqual(d["one"], 1)

    def test_many(self):
        # given
        d = PackageMetadataStub()
        # when
        d["one"] = 1
        d["one"] = 2
        # then
        self.assertEqual(d["one"], 1)

    def test_should_return_multi_keys(self):
        # given
        d = PackageMetadataStub()
        # when
        d["one"] = 1
        d["one"] = 2
        d["two"] = 3
        # then
        self.assertEqual(d.keys(), ["one", "one", "two"])

    def test_should_return_multi_items(self):
        # given
        d = PackageMetadataStub()
        # when
        d["one"] = 1
        d["one"] = 2
        d["two"] = 3
        # then
        self.assertEqual(d.items(), [("one", 1), ("one", 2), ("two", 3)])

    def test_should_return_multi_values(self):
        # given
        d = PackageMetadataStub()
        # when
        d["one"] = 1
        d["one"] = 2
        d["two"] = 3
        # then
        self.assertEqual(d.values(), [1, 2, 3])

    def test_should_init_with_dict(self):
        # given
        d = PackageMetadataStub({"one": 1, "two": 2})
        # when
        d["two"] = 3
        # then
        self.assertEqual(d.items(), [("one", 1), ("two", 2), ("two", 3)])

    def test_should_return_first_value(self):
        # given
        d = PackageMetadataStub()
        d["one"] = 1
        d["one"] = 2
        # when/then
        self.assertEqual(d.get("one"), 1)

    def test_should_return_None_if_not_found(self):
        # given
        d = PackageMetadataStub()
        # when/then
        self.assertIsNone(d.get("xxx"))

    def test_should_all_values(self):
        # given
        d = PackageMetadataStub()
        d["one"] = 1
        d["one"] = 2
        # when/then
        self.assertEqual(d.get_all("one"), [1, 2])

    def test_should_return_None_if_not_found_2(self):
        # given
        d = PackageMetadataStub()
        # when/then
        self.assertIsNone(d.get_all("xxx"))



from collections import namedtuple
from unittest import IsolatedAsyncioTestCase, mock

from packaging.requirements import Requirement
from packaging.specifiers import SpecifierSet
from packaging.version import Version

from app_utils.testing import NoSocketsTestCase

from package_monitor.core.distribution_packages import (
    DistributionPackage,
    compile_package_requirements,
    determine_system_python_version,
    gather_distribution_packages,
    gather_protected_packages_versions,
    is_marker_valid,
    is_version_in_specifiers,
    to_version_or_none,
)
from package_monitor.tests.factories import (
    DistributionPackageFactory,
    MetadataDistributionStubFactory,
    PypiFactory,
    PypiReleaseFactory,
    make_packages,
)

MODULE_PATH = "package_monitor.core.distribution_packages"


SysVersionInfo = namedtuple("SysVersionInfo", ["major", "minor", "micro"])


class TestDistributionPackage(NoSocketsTestCase):
    @mock.patch(
        MODULE_PATH + ".metadata_helpers.identify_installed_django_apps", spec=True
    )
    def test_should_create_from_importlib_distribution(self, mock_identify_django_apps):
        # given
        dist = MetadataDistributionStubFactory(
            name="Alpha",
            version="1.2.3",
            requires=["bravo>=1.0.0"],
            files=["alpha/__init__.py"],
            homepage_url="https://www.alpha.com",
        )
        mock_identify_django_apps.return_value = ["alpha_app"]
        # when
        obj = DistributionPackage.create_from_metadata_distribution(dist)
        # then
        self.assertEqual(obj.name, "Alpha")
        self.assertEqual(obj.name_normalized, "alpha")
        self.assertEqual(obj.current, "1.2.3")
        self.assertEqual(obj.latest, "")
        self.assertListEqual([str(x) for x in obj.requirements], ["bravo>=1.0.0"])
        self.assertEqual(obj.apps, ["alpha_app"])
        self.assertEqual(obj.homepage_url, "")

    def test_should_not_be_outdated(self):
        # given
        obj = DistributionPackageFactory(current="1.0.0", latest="1.0.0")
        # when/then
        self.assertFalse(obj.is_outdated())

    def test_should_be_outdated(self):
        # given
        obj = DistributionPackageFactory(current="1.0.0", latest="1.1.0")
        # when/then
        self.assertTrue(obj.is_outdated())

    def test_should_return_none_as_outdated(self):
        # given
        obj = DistributionPackageFactory(current="1.0.0", latest=None)
        # when/then
        self.assertIsNone(obj.is_outdated())

    def test_should_have_str(self):
        # given
        obj = DistributionPackageFactory(current="1.0.0", latest=None)
        # when/then
        self.assertIsInstance(str(obj), str)

    def test_should_detect_as_prerelease(self):
        # given
        obj = DistributionPackageFactory(current="1.0.0a1")
        # when/then
        self.assertTrue(obj.is_prerelease())

    def test_should_detect_not_as_prerelease(self):
        # given
        obj = DistributionPackageFactory(current="1.0.0")
        # when/then
        self.assertFalse(obj.is_prerelease())


@mock.patch(MODULE_PATH + ".importlib_metadata.distributions", spec=True)
class TestFetchRelevantPackages(NoSocketsTestCase):
    def test_should_fetch_all_packages(self, mock_distributions):
        # given
        dist_alpha = MetadataDistributionStubFactory(name="alpha")
        dist_bravo = MetadataDistributionStubFactory(
            name="bravo", requires=["alpha>=1.0.0"]
        )
        mock_distributions.return_value = [dist_alpha, dist_bravo]
        # when
        result = gather_distribution_packages()
        # then
        self.assertSetEqual({"alpha", "bravo"}, set(result.keys()))

    def test_should_ignore_corrupt_package(self, mock_distributions):
        dist_alpha = MetadataDistributionStubFactory(name="alpha")
        bad_dist = mock.Mock()
        bad_dist.metadata = {}
        bad_dist.name = "Bad Boy"
        mock_distributions.return_value = [bad_dist, dist_alpha]
        # when
        result = gather_distribution_packages()
        # then
        self.assertSetEqual({"alpha"}, set(result.keys()))


class TestCompilePackageRequirements(NoSocketsTestCase):
    def test_should_compile_requirements(self):
        # given
        dist_alpha = DistributionPackageFactory(name="alpha")
        dist_bravo = DistributionPackageFactory(name="bravo", requires=["alpha>=1.0.0"])
        packages = make_packages(dist_alpha, dist_bravo)
        # when
        with mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_CUSTOM_REQUIREMENTS", []):
            result = compile_package_requirements(packages)
        # then
        expected = {"alpha": {"bravo": SpecifierSet(">=1.0.0")}}
        self.assertDictEqual(expected, result)

    def test_should_include_requirements_from_settings(self):
        # given
        dist_alpha = DistributionPackageFactory(name="alpha")
        dist_bravo = DistributionPackageFactory(name="bravo", requires=["alpha>=1.0.0"])
        packages = make_packages(dist_alpha, dist_bravo)
        # when
        with mock.patch(
            MODULE_PATH + ".PACKAGE_MONITOR_CUSTOM_REQUIREMENTS", ["alpha>2"]
        ):
            result = compile_package_requirements(packages)
        # then
        expected = {
            "alpha": {"bravo": SpecifierSet(">=1.0.0"), "CUSTOM": SpecifierSet(">2")}
        }
        self.assertDictEqual(expected, result)

    def test_should_ignore_invalid_requirements_in_setting(self):
        # given
        dist_alpha = DistributionPackageFactory(name="alpha")
        dist_bravo = DistributionPackageFactory(name="bravo", requires=["alpha>=1.0.0"])
        packages = make_packages(dist_alpha, dist_bravo)
        # when
        with mock.patch(
            MODULE_PATH + ".PACKAGE_MONITOR_CUSTOM_REQUIREMENTS", ["alpha>2", "x!"]
        ):
            result = compile_package_requirements(packages)
        # then
        expected = {
            "alpha": {"bravo": SpecifierSet(">=1.0.0"), "CUSTOM": SpecifierSet(">2")}
        }
        self.assertDictEqual(expected, result)

    def test_should_ignore_invalid_requirements(self):
        # given
        dist_alpha = DistributionPackageFactory(name="alpha")
        dist_bravo = DistributionPackageFactory(name="bravo", requires=["alpha>=1.0.0"])
        dist_charlie = DistributionPackageFactory(name="charlie", requires=["123"])
        packages = make_packages(dist_alpha, dist_bravo, dist_charlie)
        # when
        with mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_CUSTOM_REQUIREMENTS", []):
            result = compile_package_requirements(packages)
        # then
        expected = {"alpha": {"bravo": SpecifierSet(">=1.0.0")}}
        self.assertDictEqual(expected, result)

    def test_should_ignore_python_version_requirements(self):
        # given
        dist_alpha = DistributionPackageFactory(name="alpha")
        dist_bravo = DistributionPackageFactory(name="bravo", requires=["alpha>=1.0.0"])
        dist_charlie = DistributionPackageFactory(
            name="charlie", requires=['alpha >= 1.0.0 ; python_version < "3.7"']
        )
        packages = make_packages(dist_alpha, dist_bravo, dist_charlie)
        # when
        with mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_CUSTOM_REQUIREMENTS", []):
            result = compile_package_requirements(packages)
        # then
        expected = {"alpha": {"bravo": SpecifierSet(">=1.0.0")}}
        self.assertDictEqual(expected, result)

    def test_should_ignore_invalid_extra_requirements(self):
        # given
        dist_alpha = DistributionPackageFactory(name="alpha")
        dist_bravo = DistributionPackageFactory(name="bravo", requires=["alpha>=1.0.0"])
        dist_charlie = DistributionPackageFactory(
            name="charlie", requires=['alpha>=1.0.0; extra == "certs"']
        )
        packages = make_packages(dist_alpha, dist_bravo, dist_charlie)
        # when
        with mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_CUSTOM_REQUIREMENTS", []):
            result = compile_package_requirements(packages)
        # then
        expected = {"alpha": {"bravo": SpecifierSet(">=1.0.0")}}
        self.assertDictEqual(expected, result)


@mock.patch(MODULE_PATH + ".fetch_project_from_pypi_async")
class TestUpdatePackagesFromPyPi(IsolatedAsyncioTestCase):
    def setUp(self) -> None:
        self.python_version = determine_system_python_version()

    async def test_should_update_packages(self, mock_fetch_data_from_pypi_async):
        # given
        dist_alpha = DistributionPackageFactory(name="alpha", current="1.0.0")
        requirements = {}
        pypi_alpha = PypiFactory(distribution=dist_alpha)
        pypi_alpha.releases["1.1.0"] = [PypiReleaseFactory()]
        pypi_alpha.info.version = "1.1.0"
        mock_fetch_data_from_pypi_async.return_value = pypi_alpha.asdict()
        # when
        await dist_alpha.update_from_pypi_async(
            session=mock.MagicMock(),
            requirements=requirements,
            protected_packages_versions={},
            system_python=self.python_version,
        )
        # then
        self.assertEqual(dist_alpha.latest, "1.1.0")
        self.assertEqual(dist_alpha.homepage_url, "https://pypi.org/project/alpha/")

    async def test_should_ignore_prereleases_when_stable(
        self, mock_fetch_data_from_pypi_async
    ):
        # given
        dist_alpha = DistributionPackageFactory(name="alpha", current="1.0.0")
        requirements = {}
        pypi_alpha = PypiFactory(distribution=dist_alpha)
        pypi_alpha.releases["1.1.0a1"] = [PypiReleaseFactory()]
        pypi_alpha.info.version = "1.1.0a1"
        mock_fetch_data_from_pypi_async.return_value = pypi_alpha.asdict()
        # when
        await dist_alpha.update_from_pypi_async(
            session=mock.MagicMock(),
            requirements=requirements,
            protected_packages_versions={},
            system_python=self.python_version,
        )

        # then
        self.assertEqual(dist_alpha.latest, "1.0.0")

    async def test_should_include_prereleases_when_prerelease(
        self, mock_fetch_data_from_pypi_async
    ):
        # given
        dist_alpha = DistributionPackageFactory(name="alpha", current="1.0.0a1")
        requirements = {}
        pypi_alpha = PypiFactory(distribution=dist_alpha)
        pypi_alpha.releases["1.0.0a2"] = [PypiReleaseFactory()]
        pypi_alpha.info.version = "1.0.0a2"
        mock_fetch_data_from_pypi_async.return_value = pypi_alpha.asdict()
        # when
        await dist_alpha.update_from_pypi_async(
            session=mock.MagicMock(),
            requirements=requirements,
            protected_packages_versions={},
            system_python=self.python_version,
        )

        # then
        self.assertEqual(dist_alpha.latest, "1.0.0a2")

    async def test_should_not_update_package_on_network_error(
        self, mock_fetch_data_from_pypi_async
    ):
        # given
        dist_alpha = DistributionPackageFactory(name="alpha", current="1.0.0")
        requirements = {}
        pypi_alpha = PypiFactory(distribution=dist_alpha)
        pypi_alpha.releases["1.1.0"] = [PypiReleaseFactory()]
        pypi_alpha.info.version = "1.1.0"
        mock_fetch_data_from_pypi_async.return_value = None
        # when
        await dist_alpha.update_from_pypi_async(
            session=mock.MagicMock(),
            requirements=requirements,
            protected_packages_versions={},
            system_python=self.python_version,
        )
        # then
        self.assertEqual(dist_alpha.latest, "")

    async def test_should_ignore_yanked_releases(self, mock_fetch_data_from_pypi_async):
        # given
        dist_alpha = DistributionPackageFactory(name="alpha", current="1.0.0")
        requirements = {}
        pypi_alpha = PypiFactory(distribution=dist_alpha)
        pypi_alpha.releases["1.1.0"] = [PypiReleaseFactory(yanked=True)]
        mock_fetch_data_from_pypi_async.return_value = pypi_alpha.asdict()
        # when
        await dist_alpha.update_from_pypi_async(
            session=mock.MagicMock(),
            requirements=requirements,
            protected_packages_versions={},
            system_python=self.python_version,
        )
        # then
        self.assertEqual(dist_alpha.latest, "1.0.0")

    async def test_should_ignore_releases_with_incompatible_python_requirement(
        self, mock_fetch_data_from_pypi_async
    ):
        # given
        dist_alpha = DistributionPackageFactory(name="alpha", current="1.0.0")
        requirements = {}
        pypi_alpha = PypiFactory(distribution=dist_alpha)
        pypi_alpha.releases["1.1.0"] = [PypiReleaseFactory(requires_python=">=3.7")]
        pypi_alpha.info.version = "1.1.0"
        mock_fetch_data_from_pypi_async.return_value = pypi_alpha.asdict()
        # when
        await dist_alpha.update_from_pypi_async(
            session=mock.MagicMock(),
            requirements=requirements,
            protected_packages_versions={},
            system_python=Version("3.6.9"),
        )
        # then
        self.assertEqual(dist_alpha.latest, "1.0.0")

    async def test_should_ignore_invalid_release_version(
        self, mock_fetch_data_from_pypi_async
    ):
        # given
        dist_alpha = DistributionPackageFactory(name="alpha", current="1.0.0")
        requirements = {}
        pypi_alpha = PypiFactory(distribution=dist_alpha)
        pypi_alpha.releases["a3"] = [PypiReleaseFactory()]
        pypi_alpha.info.version = "a3"
        mock_fetch_data_from_pypi_async.return_value = pypi_alpha.asdict()
        # when
        await dist_alpha.update_from_pypi_async(
            session=mock.MagicMock(),
            requirements=requirements,
            system_python=self.python_version,
            protected_packages_versions={},
        )
        # then
        self.assertEqual(dist_alpha.latest, "1.0.0")

    async def test_should_ignore_releases_not_matching_consolidated_requirements(
        self, mock_fetch_data_from_pypi_async
    ):
        # given
        dist_alpha = DistributionPackageFactory(name="alpha", current="1.0.0")
        requirements = {"alpha": {"bravo": SpecifierSet("<=1.0.0")}}
        pypi_alpha = PypiFactory(distribution=dist_alpha)
        pypi_alpha.releases["1.1.0"] = [PypiReleaseFactory()]
        pypi_alpha.info.version = "1.1.0"
        mock_fetch_data_from_pypi_async.return_value = pypi_alpha.asdict()
        # when
        await dist_alpha.update_from_pypi_async(
            session=mock.MagicMock(),
            requirements=requirements,
            system_python=self.python_version,
            protected_packages_versions={},
        )
        # then
        self.assertEqual(dist_alpha.latest, "1.0.0")

    async def test_should_ignore_invalid_python_release_spec(
        self, mock_fetch_data_from_pypi_async
    ):
        # given
        dist_alpha = DistributionPackageFactory(name="alpha", current="1.0.0")
        packages = make_packages(dist_alpha)
        requirements = {}
        pypi_alpha = PypiFactory(distribution=dist_alpha)
        pypi_alpha.releases["1.1.0"] = [PypiReleaseFactory(requires_python=">=3.4.*")]
        pypi_alpha.info.version = "1.1.0"
        mock_fetch_data_from_pypi_async.return_value = pypi_alpha.asdict()
        # when
        await dist_alpha.update_from_pypi_async(
            session=mock.MagicMock(),
            requirements=requirements,
            system_python=self.python_version,
            protected_packages_versions={},
        )
        # then
        self.assertEqual(packages["alpha"].latest, "1.0.0")

    @mock.patch(MODULE_PATH + ".fetch_pypi_releases")
    async def test_should_ignore_updates_with_non_matching_requirements_1(
        self, mock_fetch_pypi_releases, mock_fetch_data_from_pypi_async
    ):
        # given
        dist_alpha = DistributionPackageFactory(name="alpha", current="1.0.0")
        requirements = {"bravo": {"alpha": SpecifierSet(">0.1.0")}}
        pypi_alpha = PypiFactory(distribution=dist_alpha)
        pypi_alpha.info.version = "1.2.0"
        pypi_alpha.releases["1.1.0"] = [PypiReleaseFactory()]
        pypi_alpha.releases["1.2.0"] = [PypiReleaseFactory()]

        pypi_alpha_1 = PypiFactory(distribution=dist_alpha)
        pypi_alpha_1.info.version = "1.1.0"

        pypi_alpha_2 = PypiFactory(distribution=dist_alpha)
        pypi_alpha_2.info.version = "1.2.0"
        pypi_alpha_2.info.requires_dist = ["invalid>1>2", "Bravo>=1.0.0"]

        mock_fetch_data_from_pypi_async.return_value = pypi_alpha.asdict()
        mock_fetch_pypi_releases.return_value = [
            pypi_alpha_1.asdict(),
            pypi_alpha_2.asdict(),
        ]

        # when
        await dist_alpha.update_from_pypi_async(
            session=mock.MagicMock(),
            requirements=requirements,
            protected_packages_versions={"bravo": Version("0.5.0")},
            system_python=self.python_version,
        )

        # then
        self.assertEqual(dist_alpha.latest, "1.1.0")

    @mock.patch(MODULE_PATH + ".fetch_pypi_releases")
    async def test_should_ignore_updates_with_non_matching_requirements_2(
        self, mock_fetch_pypi_releases, mock_fetch_data_from_pypi_async
    ):
        """Bug #15"""
        # given
        dist_alpha = DistributionPackageFactory(name="alpha", current="1.0.0")
        requirements = {"bravo": {"alpha": SpecifierSet(">0.1.0")}}
        pypi_alpha = PypiFactory(distribution=dist_alpha)
        pypi_alpha.info.version = "1.2.0"
        pypi_alpha.releases["1.1.0"] = [PypiReleaseFactory()]
        pypi_alpha.releases["1.2.0"] = [PypiReleaseFactory()]

        pypi_alpha_1 = PypiFactory(distribution=dist_alpha)
        pypi_alpha_1.info.version = "1.1.0"

        pypi_alpha_2 = PypiFactory(distribution=dist_alpha)
        pypi_alpha_2.info.version = "1.2.0"
        pypi_alpha_2.info.requires_dist = None

        mock_fetch_data_from_pypi_async.return_value = pypi_alpha.asdict()
        mock_fetch_pypi_releases.return_value = [
            pypi_alpha_1.asdict(),
            pypi_alpha_2.asdict(),
        ]

        # when
        await dist_alpha.update_from_pypi_async(
            session=mock.MagicMock(),
            requirements=requirements,
            protected_packages_versions={"bravo": Version("0.5.0")},
            system_python=self.python_version,
        )

        # then
        self.assertEqual(dist_alpha.latest, "1.2.0")

    @mock.patch(MODULE_PATH + ".fetch_pypi_releases")
    async def test_should_ignore_conflicting_dependencies_of_not_protected_packages_1(
        self, mock_fetch_pypi_releases, mock_fetch_data_from_pypi_async
    ):
        # given

        dist_alpha = DistributionPackageFactory(name="alpha", current="1.0.0")
        requirements = {"bravo": {"alpha": SpecifierSet(">0.1.0")}}
        pypi_alpha = PypiFactory(distribution=dist_alpha)
        pypi_alpha.info.version = "1.2.0"
        pypi_alpha.releases["1.1.0"] = [PypiReleaseFactory()]
        pypi_alpha.releases["1.2.0"] = [PypiReleaseFactory()]

        pypi_alpha_1 = PypiFactory(distribution=dist_alpha)
        pypi_alpha_1.info.version = "1.1.0"

        pypi_alpha_2 = PypiFactory(distribution=dist_alpha)
        pypi_alpha_2.info.version = "1.2.0"
        pypi_alpha_2.info.requires_dist = ["invalid>1>2", "Bravo>=1.0.0"]

        mock_fetch_data_from_pypi_async.return_value = pypi_alpha.asdict()
        mock_fetch_pypi_releases.return_value = [
            pypi_alpha_1.asdict(),
            pypi_alpha_2.asdict(),
        ]

        # when
        await dist_alpha.update_from_pypi_async(
            session=mock.MagicMock(),
            requirements=requirements,
            protected_packages_versions={},
            system_python=self.python_version,
        )

        # then
        self.assertEqual(dist_alpha.latest, "1.2.0")

    @mock.patch(MODULE_PATH + ".fetch_pypi_releases")
    async def test_should_ignore_dependencies_of_not_protected_packages_2(
        self, mock_fetch_pypi_releases, mock_fetch_data_from_pypi_async
    ):
        # given

        dist_alpha = DistributionPackageFactory(name="alpha", current="1.0.0")
        requirements = {"bravo": {"alpha": SpecifierSet(">0.1.0")}}
        pypi_alpha = PypiFactory(distribution=dist_alpha)
        pypi_alpha.info.version = "1.2.0"
        pypi_alpha.releases["1.1.0"] = [PypiReleaseFactory()]
        pypi_alpha.releases["1.2.0"] = [PypiReleaseFactory()]

        pypi_alpha_1 = PypiFactory(distribution=dist_alpha)
        pypi_alpha_1.info.version = "1.1.0"

        pypi_alpha_2 = PypiFactory(distribution=dist_alpha)
        pypi_alpha_2.info.version = "1.2.0"
        pypi_alpha_2.info.requires_dist = ["charlie>=1.0.0"]

        mock_fetch_data_from_pypi_async.return_value = pypi_alpha.asdict()
        mock_fetch_pypi_releases.return_value = [
            pypi_alpha_1.asdict(),
            pypi_alpha_2.asdict(),
        ]

        # when
        await dist_alpha.update_from_pypi_async(
            session=mock.MagicMock(),
            requirements=requirements,
            protected_packages_versions={"bravo": Version("0.5.0")},
            system_python=self.python_version,
        )

        # then
        self.assertEqual(dist_alpha.latest, "1.2.0")


class TestGatherProtectedPackagesVersions(NoSocketsTestCase):
    def test_should_return_protected_packages_with_versions(self):
        # given
        dist_alpha = DistributionPackageFactory(name="alpha", current="1.0.0")
        dist_bravo = DistributionPackageFactory(name="bravo", current="1.2.3")
        packages = {"alpha": dist_alpha, "bravo": dist_bravo}

        # when
        with mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_PROTECTED_PACKAGES", ["alpha"]):
            result = gather_protected_packages_versions(packages)

        # then
        self.assertDictEqual(result, {"alpha": Version("1.0.0")})

    def test_should_return_protected_packages_with_versions_and_non_canonical_names(
        self,
    ):
        # given
        dist_alpha = DistributionPackageFactory(name="alpha", current="1.0.0")
        dist_bravo = DistributionPackageFactory(name="bravo", current="1.2.3")
        packages = {"alpha": dist_alpha, "bravo": dist_bravo}

        # when
        with mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_PROTECTED_PACKAGES", ["Alpha"]):
            result = gather_protected_packages_versions(packages)

        # then
        self.assertDictEqual(result, {"alpha": Version("1.0.0")})

    def test_should_return_empty_when_no_protected_packages(self):
        # given
        dist_alpha = DistributionPackageFactory(name="alpha", current="1.0.0")
        packages = {"alpha": dist_alpha}

        # when
        with mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_PROTECTED_PACKAGES", []):
            result = gather_protected_packages_versions(packages)

        # then
        self.assertDictEqual(result, {})

    def test_should_return_empty_when_no_packages(self):
        # given
        packages = {}

        # when
        with mock.patch(MODULE_PATH + ".PACKAGE_MONITOR_PROTECTED_PACKAGES", ["alpha"]):
            result = gather_protected_packages_versions(packages)

        # then
        self.assertDictEqual(result, {})


class TestIsRequirementValid(NoSocketsTestCase):
    def test_should_report_correctly(self):
        cases = [
            ("alpha>1", True),
            ('alpha>1; python_version>"3.0"', True),
            ('alpha>1; python_version<"3.0"', False),
        ]
        for s, expected in cases:
            with self.subTest(case=s):
                r = Requirement(s)
                self.assertIs(is_marker_valid(r), expected)


class TestToVersionOrNone(NoSocketsTestCase):
    def test_should_report_correctly(self):
        cases = [
            ("1.0.0", Version("1.0.0")),
            ("invalid", None),
            ("1.0.0alpha1", None),
        ]
        for s, expected in cases:
            with self.subTest(case=s):
                self.assertEqual(to_version_or_none(s), expected)


class TestIsVersionInSpecifiers(NoSocketsTestCase):
    def test_should_report_correctly(self):
        cases = [
            (Version("1.0.0"), SpecifierSet(">=1.0.0"), True),
            (Version("0.1.0"), SpecifierSet(">=1.0.0"), False),
            (Version("0.1.0"), SpecifierSet(""), True),
        ]
        for v, s, expected in cases:
            with self.subTest(case=s):
                self.assertEqual(is_version_in_specifiers(v, s), expected)



# Generated by Django 4.0.7 on 2022-09-12 15:03

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("package_monitor", "0002_add_editable_and_refactor_json_fields"),
    ]

    operations = [
        migrations.AddField(
            model_name="distribution",
            name="latest_notified_version",
            field=models.CharField(
                default="",
                help_text="Latest version that has been notified",
                max_length=64,
            ),
        ),
        migrations.AlterField(
            model_name="distribution",
            name="latest_version",
            field=models.CharField(
                default="",
                help_text="Latest version available for this package",
                max_length=64,
            ),
        ),
    ]



"""Managers for Package Monitor."""

from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Set

from packaging.version import parse as version_parse

from django.db import models

from allianceauth.services.hooks import get_extension_logger
from app_utils.allianceauth import notify_admins
from app_utils.logging import LoggerAddTag

from . import __title__
from .app_settings import (
    PACKAGE_MONITOR_EXCLUDE_PACKAGES,
    PACKAGE_MONITOR_INCLUDE_PACKAGES,
    PACKAGE_MONITOR_SHOW_ALL_PACKAGES,
    PACKAGE_MONITOR_SHOW_EDITABLE_PACKAGES,
)
from .core.distribution_packages import (
    DistributionPackage,
    compile_package_requirements,
    gather_distribution_packages,
    update_packages_from_pypi,
)

if TYPE_CHECKING:
    from .models import Distribution

TERMINAL_MAX_LINE_LENGTH = 4095

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class DistributionQuerySet(models.QuerySet):
    """QuerySets for Distribution."""

    def outdated_count(self) -> int:
        """Return count of outdated packages."""
        return self.filter(is_outdated=True).count()

    def build_install_command(self) -> str:
        """Build install command from all distribution packages in this query."""
        result = "pip install"
        for dist in self.exclude(latest_version=""):
            version_string = dist.pip_install_version
            if len(result) + len(version_string) + 1 > TERMINAL_MAX_LINE_LENGTH:
                break
            result = f"{result} {version_string}"
        return result

    def filter_visible(self) -> models.QuerySet:
        """Filter to include visible packages only based on current settings."""
        if PACKAGE_MONITOR_SHOW_ALL_PACKAGES:
            qs = self.all()
        else:
            qs = self.filter(has_installed_apps=True)
        if PACKAGE_MONITOR_INCLUDE_PACKAGES:
            qs |= self.filter(name__in=PACKAGE_MONITOR_INCLUDE_PACKAGES)
        if PACKAGE_MONITOR_EXCLUDE_PACKAGES:
            qs = qs.exclude(name__in=PACKAGE_MONITOR_EXCLUDE_PACKAGES)
        if not PACKAGE_MONITOR_SHOW_EDITABLE_PACKAGES:
            qs = qs.exclude(is_editable=True)
        return qs

    def names(self) -> Set[str]:
        """Return QS as set of names."""
        return set(self.values_list("name", flat=True))


class DistributionManagerBase(models.Manager):
    """Manager for Distribution."""

    def update_all(self) -> int:
        """Update the list of relevant distribution packages in the database."""
        logger.info(
            f"Started refreshing approx. {self.count()} distribution packages..."
        )
        packages = gather_distribution_packages()
        requirements = compile_package_requirements(packages)
        update_packages_from_pypi(packages, requirements)
        self._save_packages(packages=packages, requirements=requirements)
        packages_count = len(packages)
        logger.info(f"Completed refreshing {packages_count} distribution packages")
        return packages_count

    def _save_packages(
        self, packages: Dict[str, DistributionPackage], requirements: dict
    ) -> None:
        """Save the given package information into the model."""

        for package_name, package in packages.items():
            logger.debug("Updating package: %s", package)
            if package_name in requirements:
                used_by = [
                    {
                        "name": package_name,
                        "homepage_url": (
                            packages[package_name].homepage_url
                            if packages.get(package_name)
                            else ""
                        ),
                        "requirements": [str(obj) for obj in package_requirements],
                    }
                    for package_name, package_requirements in requirements[
                        package_name
                    ].items()
                ]
            else:
                used_by = []

            defaults = {
                "apps": sorted(package.apps, key=str.casefold),
                "used_by": used_by,
                "installed_version": package.current,
                "latest_version": package.latest,
                "is_outdated": package.is_outdated(),
                "is_editable": package.is_editable,
                "description": package.summary,
                "website_url": package.homepage_url,
            }
            self.update_or_create(name=package.name, defaults=defaults)

        package_names = {obj.name for obj in packages.values()}
        self.exclude(name__in=package_names).delete()

    def send_update_notification(
        self: models.QuerySet[Distribution],
        show_editable: bool,
        should_repeat: bool = False,
    ) -> None:
        """Send notification to admins to inform about new versions."""
        selected = self._filter_dist_to_notify(
            show_editable=show_editable, should_repeat=should_repeat
        )
        if selected:
            self._send_update_notification(selected)

    def _filter_dist_to_notify(
        self: models.QuerySet[Distribution],
        show_editable: bool,
        should_repeat: bool,
    ) -> List[Distribution]:
        selected = []
        for dist in self.order_by("name"):
            if dist.is_editable and not show_editable:
                continue
            if not dist.installed_version or not dist.latest_version:
                continue
            installed = version_parse(dist.installed_version)
            latest = version_parse(dist.latest_version)
            if latest <= installed:
                continue
            if not should_repeat and dist.latest_notified_version:
                notified = version_parse(dist.latest_notified_version)
                if notified >= latest:
                    continue
            selected.append(dist)
        return selected

    def _send_update_notification(self, distributions: List[Distribution]):
        count = len(distributions)
        update_text = f"{count} update" if count == 1 else f"{count} updates"
        title = f"Package Monitor: {update_text} available"
        update_text = "is one update" if count == 1 else f"are {count} updates"
        message = f"There {update_text} available:\n"
        for dist in distributions:
            message += (
                f"- {dist.name} {dist.installed_version} => {dist.latest_version}\n"
            )

        notify_admins(message=message, title=title)

        for dist in distributions:
            dist.latest_notified_version = dist.latest_version
            dist.save(update_fields={"latest_notified_version"})


DistributionManager = DistributionManagerBase.from_queryset(DistributionQuerySet)



"""Commands for Package Monitor."""

from django.core.management.base import BaseCommand

from package_monitor import __title__, __version__
from package_monitor.models import Distribution


class Command(BaseCommand):
    help = "Refreshes all data about distribution packages"

    def handle(self, *args, **options):
        self.stdout.write(f"*** {__title__} v{__version__} - Refresh Distributions ***")
        package_count = Distribution.objects.count()
        outdated_count = Distribution.objects.filter_visible().outdated_count()
        self.stdout.write(
            f"Started to refresh data for currently {package_count} distribution packages. "
            f"With {outdated_count} package(s) currently showing as outdated."
        )
        self.stdout.write("This can take a minute...Please wait")
        package_count = Distribution.objects.update_all()
        outdated_count = Distribution.objects.filter_visible().outdated_count()
        self.stdout.write(
            self.style.SUCCESS(
                f"Completed refreshing data for {package_count} distribution packages. "
                f"Identified {outdated_count} outdated package(s)."
            )
        )



"""Settings for Package Monitor."""

from app_utils.app_settings import clean_setting

PACKAGE_MONITOR_CUSTOM_REQUIREMENTS = clean_setting(
    "PACKAGE_MONITOR_CUSTOM_REQUIREMENTS", default_value=[]
)
"""List of custom requirements that all potential updates are checked against.
Example: ["gunicorn<20"]
"""

PACKAGE_MONITOR_EXCLUDE_PACKAGES = clean_setting(
    "PACKAGE_MONITOR_EXCLUDE_PACKAGES", default_value=[]
)
"""Names of distribution packages to be excluded."""


PACKAGE_MONITOR_INCLUDE_PACKAGES = clean_setting(
    "PACKAGE_MONITOR_INCLUDE_PACKAGES", default_value=[]
)
"""Names of additional distribution packages to be monitored."""


PACKAGE_MONITOR_NOTIFICATIONS_ENABLED = clean_setting(
    "PACKAGE_MONITOR_NOTIFICATIONS_ENABLED", False
)
"""Whether to notify when an update is available
for a currently installed distribution package.
"""

PACKAGE_MONITOR_NOTIFICATIONS_REPEAT = clean_setting(
    "PACKAGE_MONITOR_NOTIFICATIONS_REPEAT", False
)
"""Whether to repeat notifying about the same updates."""

PACKAGE_MONITOR_NOTIFICATIONS_SCHEDULE = clean_setting(
    "PACKAGE_MONITOR_NOTIFICATIONS_SCHEDULE", ""
)
"""When to send notifications about updates.
If not set, update notifications can be send every time the regular task runs.

The schedule can be defined in natural language. Examples: "every day at 10:00",
"every saturday at 18:00", "every first saturday every month at 15:00".
For more information about the syntax please see: [recurrent package](https://github.com/kvh/recurrent)
"""

PACKAGE_MONITOR_NOTIFICATIONS_MAX_DELAY = clean_setting(
    "PACKAGE_MONITOR_NOTIFICATIONS_MAX_DELAY", 5400
)
"""Maximum delay in seconds between the scheduled event for firing a notification
and the time the notification is issued.

This value should be synchronized with the timing of the recurring task.
"""

PACKAGE_MONITOR_SHOW_ALL_PACKAGES = clean_setting(
    "PACKAGE_MONITOR_SHOW_ALL_PACKAGES", True
)
"""Whether to show all distribution packages,
as opposed to only showing packages that contain Django apps.
"""

PACKAGE_MONITOR_SHOW_EDITABLE_PACKAGES = clean_setting(
    "PACKAGE_MONITOR_SHOW_EDITABLE_PACKAGES", False
)
"""Whether to show distribution packages installed as editable.

Since version information about editable packages is often outdated,
this type of packages are not shown by default.
"""


PACKAGE_MONITOR_PROTECTED_PACKAGES = clean_setting(
    "PACKAGE_MONITOR_PROTECTED_PACKAGES", ["allianceauth", "django"]
)
"""Names of protected packages.

Updates can include requirements for updating other packages,
which can potentially break the current AA installation.

For example: You have Django 4.2 installed
and an update to a package requires Django 5 or higher.
Then installing that package may break your installation.

When enabled Package Monitor will not show updates,
which would cause an indirect update of a protected package.

And empty list disables this feature.
"""



from collections import defaultdict
from collections.abc import MutableMapping
from dataclasses import asdict, dataclass, field
from typing import Dict, List, Optional

from importlib_metadata import PackagePath


class DjangoAppConfigStub:
    class ModuleStub:
        def __init__(self, file: str) -> None:
            self.__file__ = file

    def __init__(self, name: str, file: str) -> None:
        self.name = name
        self.module = self.ModuleStub(file)


@dataclass
class PypiUrl:
    url: str
    # incomplete


@dataclass
class PypiRelease:
    comment_text: str
    yanked: bool
    requires_python: str = ""
    # yanked_reason: str = None
    # incomplete


@dataclass
class PypiInfo:
    name: str
    version: str
    description: str = ""
    home_page: str = ""
    project_url: str = ""
    requires_dist: List[str] = field(default=list)
    requires_python: str = ""
    # summary: str
    # author: str
    # author_email: str
    # license: str
    # yanked: bool
    # yanked_reason: str = None
    # maintainer: str = None
    # maintainer_email: str = None
    # ...


@dataclass
class Pypi:
    info: PypiInfo
    last_serial: int
    releases: Dict[str, PypiRelease]
    urls: List[PypiUrl]

    def asdict(self) -> dict:
        return asdict(self)


class PackageMetadataStub(MutableMapping):
    """A PackageMetadata stub implementing the interface
    for importlib_metadata.PackageMetadata.

    This is a dict-like type, which allows multiple values per key.
    """

    def __init__(self, init_d: Optional[dict] = None):
        self._d = defaultdict(list)
        if init_d:
            for k, v in init_d.items():
                self[k] = v

    def __setitem__(self, key, value):
        self._d[key].append(value)

    def __getitem__(self, key):
        if key not in self._d:
            return None
        return self._d[key][0]

    def __delitem__(self, __key) -> None:
        raise NotImplementedError()

    def __len__(self) -> int:
        return len(self._d)

    def __iter__(self):
        return self.keys()

    def items(self):
        results = []
        for key, values in self._d.items():
            for value in values:
                results.append((key, value))
        return results

    def keys(self):
        return [o[0] for o in self.items()]

    def values(self):
        return [o[1] for o in self.items()]

    def get_all(self, key) -> Optional[list]:
        v = self._d[key]
        if not v:
            return None
        return v


class MetadataDistributionStub:
    def __init__(
        self,
        name: str,
        version: str,
        files: list,
        requires: Optional[list] = None,
        homepage_url: str = "",
        summary: str = "",
    ) -> None:
        self.metadata = PackageMetadataStub(
            {
                "Name": name,
                "Home-page": homepage_url if homepage_url != "" else "UNKNOWN",
                "Summary": summary if summary != "" else "UNKNOWN",
                "Version": version if version != "" else "UNKNOWN",
            }
        )
        self.files = [PackagePath(f) for f in files]
        self.requires = requires if requires else None
        self._files_content = {}

    @property
    def name(self):
        return self.metadata["Name"]

    @property
    def version(self):
        return self.metadata["Version"]

    def read_text(self, filename: str) -> Optional[str]:
        if filename in self._files_content:
            return self._files_content[filename]
        return ""



from unittest import mock

from packaging.requirements import Requirement

from app_utils.testing import NoSocketsTestCase

from package_monitor.core.metadata_helpers import (
    _extract_files,
    identify_installed_django_apps,
    is_distribution_editable,
    metadata_value,
    parse_requirements,
)

from ..factories import DjangoAppConfigStub, MetadataDistributionStubFactory

MODULE_PATH = "package_monitor.core.metadata_helpers"


@mock.patch(MODULE_PATH + ".os.path.isfile")
class TestIsDistributionEditable(NoSocketsTestCase):
    def test_should_not_be_editable(self, mock_isfile):
        # given
        mock_isfile.return_value = False
        obj = MetadataDistributionStubFactory(name="alpha")
        # when/then
        self.assertFalse(is_distribution_editable(obj))

    def test_should_be_editable_old_version(self, mock_isfile):
        # given
        mock_isfile.return_value = True
        obj = MetadataDistributionStubFactory(name="alpha")
        # when/then
        self.assertTrue(is_distribution_editable(obj))

    def test_should_be_editable_pep660(self, mock_isfile):
        # given
        mock_isfile.return_value = False

        obj = MetadataDistributionStubFactory(name="alpha")
        obj._files_content = {
            "direct_url.json": '{"dir_info": {"editable": true}, "url": "xxx"}'
        }
        # when/then
        self.assertTrue(is_distribution_editable(obj))

    def test_should_not_be_editable_pep660(self, mock_isfile):
        # given
        mock_isfile.return_value = False

        obj = MetadataDistributionStubFactory(name="alpha")
        obj._files_content = {
            "direct_url.json": '{"dir_info": {"editable": false}, "url": "xxx"}'
        }
        # when/then
        self.assertFalse(is_distribution_editable(obj))


class TestExtractFiles(NoSocketsTestCase):
    def test_should_return_empty_list_when_no_files_match(self):
        # given
        dist = MetadataDistributionStubFactory()
        # when/then
        self.assertListEqual(_extract_files(dist, "__init__.py"), [])

    def test_should_return_matching_files(self):
        # given
        dist = MetadataDistributionStubFactory(
            files=["/alpha/xx.py", "/bravo/green/__init__.py", "/charlie/yy.py"]
        )
        # when/then
        self.assertListEqual(
            _extract_files(dist, "__init__.py"), ["/bravo/green/__init__.py"]
        )


@mock.patch(MODULE_PATH + ".django_apps", spec=True)
class TestIdentifyInstalledDjangoApps(NoSocketsTestCase):
    def test_should_identify_app(self, mock_django_apps):
        # given
        dist = MetadataDistributionStubFactory(
            files=["alpha/__init__.py", "alpha/apps.py"]
        )
        app = DjangoAppConfigStub("alpha_app", "alpha/__init__.py")
        mock_django_apps.get_app_configs.return_value = [app]
        # when
        result = identify_installed_django_apps(dist)
        # then
        self.assertListEqual(result, ["alpha_app"])

    def test_should_not_identify_app(self, mock_django_apps):
        # given
        dist = MetadataDistributionStubFactory(files=["alpha/apps.py"])
        mock_django_apps.get_app_configs.return_value = []
        # when
        result = identify_installed_django_apps(dist)
        # then
        self.assertListEqual(result, [])

    def test_should_handle_no_module_file(self, mock_django_apps):
        # given
        dist = MetadataDistributionStubFactory(
            files=["alpha/__init__.py", "alpha/apps.py"]
        )
        app = DjangoAppConfigStub("alpha_app", None)
        mock_django_apps.get_app_configs.return_value = [app]
        # when
        result = identify_installed_django_apps(dist)
        # then
        self.assertListEqual(result, [])


class TestParseRequirements(NoSocketsTestCase):
    def test_should_parse_requirements_correctly(self):
        # given
        dist = MetadataDistributionStubFactory(requires=["bravo>=1.0.0", "alpha<5"])

        # when
        result = parse_requirements(dist)
        # then
        self.assertEqual(result, [Requirement("bravo>=1.0.0"), Requirement("alpha<5")])

    def test_should_ignore_invalid_requirements(self):
        # given
        dist = MetadataDistributionStubFactory(
            requires=["alpha<5;invalid", "bravo>=1.0.0"]
        )
        # when
        result = parse_requirements(dist)
        # then
        self.assertEqual(result, [Requirement("bravo>=1.0.0")])


# class TestDetermineHomePageUrl(NoSocketsTestCase):
#     def test_should_identify_homepage_old_style(self):
#         # given
#         dist = ImportlibDistributionStubFactory(homepage_url="my-homepage-url")
#         # when
#         url = _determine_homepage_url(dist)
#         # then
#         self.assertEqual(url, "my-homepage-url")

#     def test_should_identify_homepage_pep_621_style(self):
#         # given
#         dist = ImportlibDistributionStubFactory(homepage_url="")
#         for v in [
#             "Documentation, other-url",
#             "Homepage, my-homepage-url",
#             "Issues, other-url",
#         ]:
#             dist.metadata["Project-URL"] = v
#         # when
#         url = _determine_homepage_url(dist)
#         # then
#         self.assertEqual(url, "my-homepage-url")

#     def test_should_identify_homepage_pep_621_style_other_case(self):
#         # given
#         dist = ImportlibDistributionStubFactory(homepage_url="")
#         for v in [
#             "Documentation, other-url",
#             "homepage, my-homepage-url",
#             "Issues, other-url",
#         ]:
#             dist.metadata["Project-URL"] = v
#         # when
#         url = _determine_homepage_url(dist)
#         # then
#         self.assertEqual(url, "my-homepage-url")

#     def test_should_return_empty_string_when_no_url_found_with_pep_621(self):
#         # given
#         dist = ImportlibDistributionStubFactory(homepage_url="")
#         for v in ["Documentation, other-url", "Issues, other-url"]:
#             dist.metadata["Project-URL"] = v
#         # when
#         url = _determine_homepage_url(dist)
#         # then
#         self.assertEqual(url, "")


class TestDistMetadataValue(NoSocketsTestCase):
    def test_should_return_value_when_exists(self):
        # given
        dist = MetadataDistributionStubFactory(name="Alpha")
        # when/then
        self.assertEqual(metadata_value(dist, "Name"), "Alpha")

    def test_should_return_empty_string_when_prop_does_not_exist(self):
        # given
        dist = MetadataDistributionStubFactory(name="Alpha")
        # when/then
        self.assertEqual(metadata_value(dist, "XXX"), "")

    def test_should_return_name(self):
        # given
        dist = MetadataDistributionStubFactory(name="Alpha")
        # when/then
        self.assertEqual(metadata_value(dist, "Name"), "Alpha")

    def test_should_return_empty_string_when_value_is_undefined(self):
        # given
        dist = MetadataDistributionStubFactory(homepage_url="")
        # when/then
        self.assertEqual(metadata_value(dist, "Home-page"), "")

    def test_should_return_empty_string_when_value_is_none(self):
        # given
        dist = MetadataDistributionStubFactory(homepage_url=None)
        # when/then
        self.assertEqual(metadata_value(dist, "Home-page"), "")



from typing import Generic, TypeVar

import factory
import factory.fuzzy
from packaging.requirements import Requirement

from package_monitor.core.distribution_packages import DistributionPackage
from package_monitor.models import Distribution

from .stubs import MetadataDistributionStub, Pypi, PypiInfo, PypiRelease, PypiUrl

T = TypeVar("T")

faker = factory.faker.faker.Faker()


class BaseMetaFactory(Generic[T], factory.base.FactoryMetaClass):
    def __call__(cls, *args, **kwargs) -> T:
        return super().__call__(*args, **kwargs)


class PypiReleaseFactory(factory.Factory):
    class Meta:
        model = PypiRelease

    comment_text = factory.faker.Faker("sentence")
    yanked = False


class PypiUrlFactory(factory.Factory):
    class Meta:
        model = PypiUrl

    url = factory.faker.Faker("url")


class PypiInfoFactory(factory.Factory):
    class Meta:
        model = PypiInfo
        exclude = ("requirements",)

    project_url = factory.LazyAttribute(lambda o: f"https://pypi.org/project/{o.name}/")
    requires_dist = factory.LazyAttribute(
        lambda o: [str(obj) for obj in o.requirements]
    )
    requires_python = "~=3.7"


class PypiFactory(factory.Factory, metaclass=BaseMetaFactory[Pypi]):
    """A data object on PyPI. Create from DistributionPackage"""

    class Meta:
        model = Pypi
        exclude = ("distribution",)

    info = factory.LazyAttribute(
        lambda o: PypiInfoFactory(
            name=o.distribution.name,
            version=o.distribution.current,
            description=o.distribution.summary,
            home_page=o.distribution.homepage_url,
            requirements=o.distribution.requirements,
        )
    )
    last_serial = factory.fuzzy.FuzzyInteger(1_000_000, 10_000_000)
    releases = factory.LazyAttribute(
        lambda o: {o.distribution.current: [PypiReleaseFactory()]}
    )
    urls = factory.LazyAttribute(lambda o: [PypiUrlFactory()])


class MetadataDistributionStubFactory(
    factory.Factory, metaclass=BaseMetaFactory[MetadataDistributionStub]
):
    class Meta:
        model = MetadataDistributionStub

    name = factory.Faker("last_name")
    # files = ["dummy_1/file_1.py", "dummy_1/__init__.py"]
    homepage_url = factory.Faker("url")
    summary = factory.Faker("sentence")

    @factory.lazy_attribute
    def version(self):
        int_fuzzer = factory.fuzzy.FuzzyInteger(0, 20)
        major = int_fuzzer.fuzz()
        minor = int_fuzzer.fuzz()
        patch = int_fuzzer.fuzz()
        return f"{major}.{minor}.{patch}"

    @factory.lazy_attribute
    def files(self):
        path = faker.words(1)[0]
        files = faker.words(3)
        return [f"{path}/{file}.py" for file in files]

    @classmethod
    def _adjust_kwargs(cls, **kwargs):
        requires = kwargs.get("requires")
        if requires and (
            isinstance(requires, str) or not hasattr(requires, "__iter__")
        ):
            raise RuntimeError(f"requires need to be an iterable: {requires}")
        return kwargs


class DistributionPackageFactory(
    factory.Factory, metaclass=BaseMetaFactory[DistributionPackage]
):
    class Meta:
        model = DistributionPackage
        exclude = ("requires",)

    requires = ""  # excluded

    name = factory.Faker("last_name")
    is_editable = False
    # latest = factory.LazyAttribute(lambda o: o.current)
    homepage_url = factory.Faker("url")
    summary = factory.Faker("sentence")

    @factory.lazy_attribute
    def current(self):
        int_fuzzer = factory.fuzzy.FuzzyInteger(0, 20)
        major = int_fuzzer.fuzz()
        minor = int_fuzzer.fuzz()
        patch = int_fuzzer.fuzz()
        return f"{major}.{minor}.{patch}"

    @factory.lazy_attribute
    def requirements(self):
        if self.requires:
            return [Requirement(obj) for obj in self.requires]
        return []


class DistributionFactory(
    factory.django.DjangoModelFactory, metaclass=BaseMetaFactory[Distribution]
):
    class Meta:
        model = Distribution
        django_get_or_create = ("name",)

    name = factory.Faker("last_name")
    description = factory.Faker("paragraph")
    latest_version = factory.LazyAttribute(lambda o: o.installed_version)
    is_outdated = False
    website_url = factory.Faker("uri")

    @factory.lazy_attribute
    def installed_version(self):
        int_fuzzer = factory.fuzzy.FuzzyInteger(0, 20)
        major = int_fuzzer.fuzz()
        minor = int_fuzzer.fuzz()
        patch = int_fuzzer.fuzz()
        return f"{major}.{minor}.{patch}"


def make_packages(*packages: DistributionPackage):
    return {obj.name_normalized: obj for obj in packages}



import datetime as dt
from unittest.mock import patch

from django.test import TestCase, override_settings
from django.utils.timezone import now

from package_monitor import tasks

MODULE_PATH = "package_monitor.tasks"
UTC = dt.timezone.utc


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
@patch(MODULE_PATH + ".cache.get", spec=True)
@patch(MODULE_PATH + ".schedule.is_notification_due")
class TestShouldSendNotifications(TestCase):
    def test_should_send_notification_when_due(self, is_notification_due, cache_get):
        # given
        is_notification_due.return_value = True
        last_report = (dt.datetime(2024, 7, 10, 10, 0, 0, tzinfo=UTC),)
        cache_get.return_value = last_report
        # when
        with patch(MODULE_PATH + ".PACKAGE_MONITOR_NOTIFICATIONS_ENABLED", True), patch(
            MODULE_PATH + ".PACKAGE_MONITOR_NOTIFICATIONS_SCHEDULE", "my schedule"
        ), patch(MODULE_PATH + ".PACKAGE_MONITOR_NOTIFICATIONS_MAX_DELAY", 42):
            result = tasks._should_send_notifications()
        # then
        self.assertTrue(result)
        _, kwargs = is_notification_due.call_args
        self.assertEqual(kwargs["schedule_text"], "my schedule")
        self.assertEqual(kwargs["last_report"], last_report)
        self.assertEqual(kwargs["max_delay"], 42)

    def test_should_not_send_notification_when_not_due(
        self, is_notification_due, cache_get
    ):
        # given
        is_notification_due.return_value = False
        last_report = (dt.datetime(2024, 7, 10, 10, 0, 0, tzinfo=UTC),)
        cache_get.return_value = last_report
        # when
        with patch(MODULE_PATH + ".PACKAGE_MONITOR_NOTIFICATIONS_ENABLED", True):
            result = tasks._should_send_notifications()
        # then
        self.assertFalse(result)

    def test_should_not_send_notification_when_disabled(
        self, is_notification_due, cache_get
    ):
        # when
        with patch(MODULE_PATH + ".PACKAGE_MONITOR_NOTIFICATIONS_ENABLED", False):
            result = tasks._should_send_notifications()
        # then
        self.assertFalse(result)


@patch(MODULE_PATH + ".cache.set", spec=True)
@patch(MODULE_PATH + ".Distribution.objects.send_update_notification")
class TestSendUpdateNotifications(TestCase):
    def test_can_send_notifications(self, send_update_notification, cache_set):
        # when
        with patch(
            MODULE_PATH + ".PACKAGE_MONITOR_SHOW_EDITABLE_PACKAGES", False
        ), patch(MODULE_PATH + ".PACKAGE_MONITOR_NOTIFICATIONS_REPEAT", False):
            tasks.send_update_notification()
        # then
        self.assertTrue(send_update_notification.called)
        _, kwargs = send_update_notification.call_args
        self.assertFalse(kwargs["should_repeat"])
        self.assertFalse(kwargs["show_editable"])
        self.assertTrue(cache_set.called)
        _, kwargs = cache_set.call_args
        self.assertAlmostEqual(kwargs["value"], now(), delta=dt.timedelta(seconds=10))


@patch(MODULE_PATH + ".Distribution")
@patch(MODULE_PATH + "._should_send_notifications")
@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestUpdateDistributions(TestCase):
    def test_should_update_and_notify(self, should_send_notifications, Distribution):
        # given
        should_send_notifications.return_value = True
        # when
        tasks.update_distributions()
        # then
        self.assertTrue(Distribution.objects.update_all.called)
        self.assertTrue(Distribution.objects.send_update_notification.called)

    def test_should_update_and_not_notify(
        self, should_send_notifications, Distribution
    ):
        # given
        should_send_notifications.return_value = False
        # when
        tasks.update_distributions()
        # then
        self.assertTrue(Distribution.objects.update_all.called)
        self.assertFalse(Distribution.objects.send_update_notification.called)



"""Module schedule calculates events for the update notification schedule."""

import datetime as dt
from typing import Optional

import pytz
from dateutil import rrule
from recurrent.event_parser import RecurringEvent

from django.utils.timezone import now


def is_notification_due(
    schedule_text: str, max_delay: int, last_report: Optional[dt.datetime]
) -> bool:
    """Reports whether a new update notification is due.

    Args:
        - schedule_text: Frequency in natural language
        - max_delay: how much seconds can pass after the scheduled event to still fire
        - last_report: Date of the last report or None
    """
    if not schedule_text:
        return True
    r = RecurringEvent()
    r.parse(schedule_text)
    now_2 = _aware2native(now())
    # occurrences needs to start way in the past
    # so we can always determine the previous occurrence
    r.dtstart = now_2 - dt.timedelta(days=30 * 6)
    rule = rrule.rrulestr(r.get_RFC_rrule())
    previous = rule.before(now_2)
    previous_latest = previous + dt.timedelta(seconds=max_delay)
    if last_report:
        last_report_2 = _aware2native(last_report)
        if previous <= last_report_2 <= previous_latest:
            return False
    return previous <= now_2 <= previous_latest


def _aware2native(d: dt.datetime) -> dt.datetime:
    return d.astimezone(pytz.utc).replace(tzinfo=None)
